from typing import Optional

from formsg.constants import PUBLIC_KEY_PRODUCTION, PUBLIC_KEY_STAGING
from formsg.crypto import Crypto, DecryptParams
from formsg.webhook import Webhook


class FormSdk(object):
    # TODO: type(mode) == Literal
    def __init__(self, mode: str, webhook_secret_key: Optional[str] = None):
        self.mode = mode
        self.public_key: str
        if self.mode == "STAGING":
            self.public_key = PUBLIC_KEY_STAGING
        elif self.mode == "PRODUCTION":
            self.public_key = PUBLIC_KEY_PRODUCTION
        else:  # default to prod
            self.public_key = PUBLIC_KEY_PRODUCTION

        self.crypto = Crypto(self.public_key)
        self.webhooks = Webhook(self.public_key, webhook_secret_key)



from typing import Dict, Union, cast

from formsg.schemas.verification import (
    HeaderSignatureSchema,
    VerificationSignatureSchema,
)


def _extract_key_values(s: str) -> dict:
    tokens = s.split(",")
    d: Dict[str, Union[str, int]] = dict()
    for token in tokens:
        k, v = token.split("=", maxsplit=1)
        d[k] = v
    return d


def parse_signature_header(header: str) -> HeaderSignatureSchema:
    parsed_signature = cast(HeaderSignatureSchema, {})
    parsed_signature = _extract_key_values(header)  # type: ignore
    parsed_signature["t"] = int(parsed_signature["t"])
    return parsed_signature


def parse_verification_signature(signature: str) -> VerificationSignatureSchema:
    parsed_signature = cast(VerificationSignatureSchema, {})
    parsed_signature = _extract_key_values(signature)  # type: ignore
    parsed_signature["t"] = int(parsed_signature["t"])
    return parsed_signature



import base64
import json
import logging
from typing import Union, cast

import requests
from nacl.exceptions import CryptoError
from nacl.public import Box, PrivateKey, PublicKey

from formsg.exceptions import AttachmentDecryptionException, MissingPublicKeyException
from formsg.schemas.crypto import (
    DecryptedContent,
    DecryptedContentAndAttachments,
    DecryptedFile,
    DecryptParams,
)
from formsg.util.crypto import (
    are_attachment_field_ids_valid,
    convert_encrypted_attachment_to_file_content,
    decrypt_content,
    verify_signed_message,
)
from formsg.util.validate import determine_is_form_fields

logger = logging.getLogger(__name__)


class Crypto(object):
    def __init__(self, signing_public_key: str):
        self.signing_public_key = signing_public_key

    def decrypt(
        self, form_secret_key: str, decrypt_params: DecryptParams
    ) -> Union[DecryptedContent, None]:
        """
        Decrypts an encrypted submission and returns it.
        :param: form_secret_key The base-64 secret key of the form to decrypt with.
        :param: decrypt_params :class:`dict` The params containing encrypted content and information
        :param: decrypt_params.encryptedContent The encrypted content encoded with base-64.
        :param: decrypt_params.version The version of the payload. Used to determine the decryption process to decrypt the content with.
        :param: decrypt_params.verifiedContent Optional. The encrypted and signed verified content. If given, the signingPublicKey will be used to attempt to open the signed message.
        :returns: The decrypted content if successful. Else, null will be returned.
        :raises MissingPublicKeyException: if a public key is not provided when instantiating this class and is needed for verifying signed content.
        """
        try:
            if "encryptedContent" not in decrypt_params:
                logger.error("`encryptedContent` not provided")
                return None

            decrypted_bytes = decrypt_content(
                form_secret_key, decrypt_params["encryptedContent"]
            )
            if not decrypted_bytes:
                raise Exception("Failed to decrypt content")

            decrypted_object = json.loads(decrypted_bytes.decode("utf-8"))
            if not determine_is_form_fields(decrypted_object):
                print("here")
                raise Exception("Decrypted object does not fit expected shape")

            returned_object: DecryptedContent = {
                "responses": decrypted_object,
            }

            if "verifiedContent" in decrypt_params:
                decrypted_verified_object = self._decrypt_verified_content(
                    form_secret_key, decrypt_params
                )
                returned_object["verified"] = decrypted_verified_object

            return returned_object
        except MissingPublicKeyException as e:
            logger.error(e)
            raise e
        except:
            return None

    def decrypt_file(
        self, form_secret_key: str, encrypted_file_content
    ) -> Union[bytes, None]:
        """
        Decrypt the given encrypted file content.
        :param form_secret_key Secret key as a base-64 string
        :param encrypted_file_content Object returned from encryptFile function
        :param encrypted_file_content.submissionPublicKey The submission public key as a base-64 string
        :param encrypted_file_content.nonce The nonce as a base-64 string
        :param encrypted_file_content.blob The encrypted file as a Blob object

        """
        box = Box(
            PrivateKey(base64.b64decode(form_secret_key)),
            PublicKey(
                base64.b64decode(encrypted_file_content["submission_public_key"])
            ),
        )
        try:
            decrypted = box.decrypt(
                encrypted_file_content["binary"],
                base64.b64decode(encrypted_file_content["nonce"]),
            )
            return decrypted
        except CryptoError:
            logger.error("Error decrypting file")
            return None

    def decrypt_attachments(
        self, form_secret_key: str, decrypt_params: DecryptParams
    ) -> Union[DecryptedContentAndAttachments, None]:
        """
        Decrypts an encrypted submission, and also download and decrypt any attachments alongside it.
        :param form_secret_key Secret key as a base-64 string
        :param decrypt_params The params containing encrypted content and information.
        :rtype object: An object containing the decrypted submission, including attachments (if any). Or else returns null if a decryption error decrypting any part of the submission.
        :raises MissingPublicKeyException if a public key is not provided when instantiating this class and is needed for verifying signed content.
        """

        if "attachmentDownloadUrls" not in decrypt_params:
            logger.error("`attachmentDownloadUrls` param not passed")
            return None

        attachment_records = decrypt_params.get("attachmentDownloadUrls", {})
        decrypted_content = self.decrypt(form_secret_key, decrypt_params)
        if not decrypted_content:
            return None
        returned_object: DecryptedContentAndAttachments = {
            "content": decrypted_content,
            "attachments": {},
        }

        decrypted_records = {}
        filenames = {}
        for response in decrypted_content["responses"]:
            if response["fieldType"] == "attachment" and response["answer"]:
                filenames[response["_id"]] = response["answer"]

        field_ids = attachment_records.keys()  # type: ignore
        if not are_attachment_field_ids_valid(list(field_ids), filenames):
            return None

        # possible improvement: make this parallel (eg. with grequests, request_futures)
        try:
            for field_id in field_ids:
                resp = requests.get(attachment_records[field_id])  # type: ignore
                data = resp.json()
                encrypted_file = convert_encrypted_attachment_to_file_content(data)
                decrypted_file = self.decrypt_file(form_secret_key, encrypted_file)
                if not decrypted_file:
                    raise AttachmentDecryptionException()
                decrypted_record: DecryptedFile = {
                    "filename": filenames[field_id],
                    "content": decrypted_file,
                }
                decrypted_records[field_id] = decrypted_record
        except AttachmentDecryptionException:
            raise
        except Exception as e:
            logger.error(e)
            return None
        returned_object["attachments"] = decrypted_records
        return returned_object

    def _decrypt_verified_content(
        self, form_secret_key: str, decrypt_params: DecryptParams
    ):
        if not self.signing_public_key:
            raise MissingPublicKeyException(
                "Public signing key must be provided when instantiating the Crypto class in order to verify verified content"
            )
        decrypted_verified_content = decrypt_content(
            form_secret_key, decrypt_params["verifiedContent"]  # type: ignore
        )
        if not decrypted_verified_content:
            raise Exception("Failed to decrypt verified content")

        return verify_signed_message(
            decrypted_verified_content, self.signing_public_key
        )



class WebhookAuthenticateException(Exception):
    pass


class AttachmentDecryptionException(Exception):
    pass


class MissingPublicKeyException(Exception):
    pass


class MissingSecretKeyException(Exception):
    pass



def is_signature_time_valid(
    signature_time: int, submission_created_at: int, transaction_expiry: int
) -> bool:
    """
    Checks if signature was made within the given expiry range before submission was created.

    :param signature_time: in ms
    :param submission_created_at: in ms
    :rtype: :class:`bool`
    """
    max_time = submission_created_at
    min_time = max_time - transaction_expiry * 1000
    return signature_time > min_time and signature_time < max_time


def format_to_base_string(params) -> str:
    """
    Formats given data into a string for signing
    """
    [transaction_id, form_id, field_id, answer, time_] = [
        params["transactionId"],
        params["formId"],
        params["fieldId"],
        params["answer"],
        params["time"],
    ]

    return f"{transaction_id}.{form_id}.{field_id}.{answer}.{time_}"



import base64
import logging
import time
import urllib

from nacl.bindings.crypto_sign import crypto_sign, crypto_sign_BYTES
from nacl.signing import VerifyKey
from typing_extensions import TypedDict

from formsg.exceptions import WebhookAuthenticateException

logger = logging.getLogger(__name__)

SignatureHeader = TypedDict(
    "SignatureHeader",
    {"v1": str, "t": int, "s": str, "f": str},
)


def is_signature_valid(
    uri: str, signature_header: SignatureHeader, public_key: str
) -> bool:
    """
    Helper function to construct the basestring and verify the signature of an
    incoming request
    :param: uri incoming request to verify
    :param: signatureHeader the X-FormSG-Signature header to verify against
    :rtype: :class:`bool` true if verification succeeds, false otherwise
    raises {WebhookAuthenticateError} if given signature header is malformed.
    """
    [signature, epoch, submission_id, form_id] = [
        signature_header["v1"],
        signature_header["t"],
        signature_header["s"],
        signature_header["f"],
    ]
    if not epoch or not signature or not submission_id or not form_id:
        raise WebhookAuthenticateException("X-FormSG-Signature header is invalid")

    parsed_url = urllib.parse.urlparse(uri).geturl()
    base_string = f"{parsed_url}.{submission_id}.{form_id}.{epoch}"
    v_key = VerifyKey(base64.b64decode(public_key))
    try:
        _verify(v_key, base_string, signature)
        return True
    except Exception as e:
        logger.error(e)
        return False


def _verify(verify_key: VerifyKey, uri: str, signature: str) -> bytes:
    return verify_key.verify(uri.encode("utf-8"), base64.b64decode(signature))


def has_epoch_expired(epoch: int, expiry: int = 300000) -> bool:
    """
    :param epoch: time in ms
    :param expiry: time in ms
    :rtype :class:`bool` if epoch has expired
    """
    # epoch is in ms, time.time() is in sec
    difference = abs(time.time() * 1000 - epoch)
    return difference > expiry


"""
# `nacl.signing.SigningKey` doesn't allow for a 64 bit salt, so we need to use their lower level API
# ref https://github.com/pyca/pynacl/issues/419#issuecomment-377520844
"""


def sign(base_string: str, secret_key: str) -> bytes:
    combined = crypto_sign(base_string.encode("utf-8"), base64.b64decode(secret_key))
    return base64.b64encode(combined[:crypto_sign_BYTES])
